/*
 * Sneaky-Seq - Phrase Sequencer Based on TK's Clockbox (MIDI Clock Sender)
 *
 * This sequencer triggers loops in Native Instruments Kontakt.
 *
 * Each loop has an original BPM and will play at the current BPM
 * of this sequencer - this sequencer does this by transmitting midi 
 * CC 119 from 40 - 127. A custom Kontakt script reads this midi cc 
 * and sends a corresponding pitch-bend command to the loops so that 
 * the loops play at the current tempo of the sequencer.
 * 
 * The sequencer has 16 steps and six tracks per section. Only 3 tracks
 * are currently implemented for testing purposes. Scaling to 6 should
 * not be difficult and should not affect midi timing adversely.
 * 
 * ==========================================================================
 *
 * Copyright (C) 2005  Thorsten Klose (tk@midibox.org)
 * 
 * ==========================================================================
 * 
 * This file is part of a MIOS application
 *
 * This application is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This application is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with This application; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ==========================================================================
 */

/////////////////////////////////////////////////////////////////////////////
// Include files
/////////////////////////////////////////////////////////////////////////////

#include "cmios.h"
#include "pic18f452.h"

#include "main.h"
#include "mclock.h"

/////////////////////////////////////////////////////////////////////////////
// Global variables
/////////////////////////////////////////////////////////////////////////////

// status of application (see bitfield declaration in main.h)
app_flags_t app_flags;

/////////////////////////////////////////////////////////////////////////////
// Local variables
/////////////////////////////////////////////////////////////////////////////


// last ain/din/dout
unsigned char last_ain_pin;
unsigned char last_din_pin;
unsigned char last_dout_pin;

// Array for AIN to be assigned to custom CC mapping

const unsigned char pot_event_map[8][2] = {
  // Pot 1..16 send CC#0x5b (Reverb) over MIDI Channel #1..16
  {0xb0, 0x0d},   {0xb0, 0x02},   {0xb0, 0x4a},   {0xb0, 0x04}, 
  {0xb0, 0x0c},   {0xb0, 0x03},   {0xb0, 0x4b},   {0xb0, 0x0e}
}; 


// here are my song names

const unsigned char songname[47][10] = {
"Mistress ",
"BodyMilit",
"YeuxFerme",
"Tyrant   ",
"No Escape",
"AfterDark",
"Jackaha  ",
"Russian  ",
"MysticEye",
"RobinsFnk",
"FluidLife",
"Firenze  ",
"PBoysThme",
"BlackFlwr",
"Ghostz   ",
"RawMision",
"ColdWays ",
"PlsurePat",
"Transprnz",
"AstralPln",
"GlobalMis",
"LuckyBmbs",
"Material ",
"Vivez    ",
"LogicalTr",
"MinusMan ",
"AntiKnfkt",
"G-String ",
"Spiral   ",
"HeartBrkr",
"BetwnDark",
"Pravda   ",
"PussyChrm",
"CaptainsH",
"BindngFrc",
"Wake Up  ",
"Hustlers ",
"Un-Known ",
"AurientXp",
"Reply 23 ",
"New Breed",
"Metroplis",
"YourPhant",
"Love Spy ",
"Bordello ",
"WithoutTr",
"LifeSuprt"
};
	

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS after startup to initialize the 
// application
/////////////////////////////////////////////////////////////////////////////
void Init(void) __wparam
{
  // set shift register update frequency
  MIOS_SRIO_UpdateFrqSet(1); // ms

  // we need to set at least one IO shift register pair
  MIOS_SRIO_NumberSet(NUMBER_OF_SRIO);

  // debouncing value for DINs
  MIOS_SRIO_DebounceSet(DIN_DEBOUNCE_VALUE);

  // Touch sensor sensitivity *must* be 0, otherwise Port D.4 (CORE::J14) cannot be used as Clock Output
  MIOS_SRIO_TS_SensitivitySet(0);

  // set encoder speed mode of datawheel
  MIOS_ENC_SpeedSet(0, DATAWHEEL_SPEED_MODE, DATAWHEEL_SPEED_DIVIDER);

  // initialize the AIN driver
  MIOS_AIN_NumberSet(AIN_NUMBER_INPUTS);
  #if AIN_MUXED_MODE
  MIOS_AIN_Muxed();
  #else
  MIOS_AIN_UnMuxed();
  #endif
  MIOS_AIN_DeadbandSet(AIN_DEADBAND);



  // initialize the MIDI clock module (-> mclock.c)
  MCLOCK_Init();
  MCLOCK_BPMSet(120);
  MCLOCK_SongNumberSet(0);
  MCLOCK_LoadSeq(1); 

}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS in the mainloop when nothing else is to do
/////////////////////////////////////////////////////////////////////////////
void Tick(void) __wparam
{
  // this routine sends the MIDI clock (and Start/Continue/Stop) if requested
  MCLOCK_Tick();
  
}

/////////////////////////////////////////////////////////////////////////////
// This function is periodically called by MIOS. The frequency has to be
// initialized with MIOS_Timer_Set
/////////////////////////////////////////////////////////////////////////////
void Timer(void) __wparam
{
  // forward timer event to MIDI clock module (-> mclock.c)
  MCLOCK_Timer();
  
}





/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when the display content should be 
// initialized. Thats the case during startup and after a temporary message
// has been printed on the screen
/////////////////////////////////////////////////////////////////////////////

// default special character set
// created with http://lcd5x8char.midibox.org
//        ;; char #1
//	db b'00001000', b'00001100'; line 1 / 2
//	db b'00001110', b'00001111'; line 3 / 4
//	db b'00001111', b'00001110'; line 5 / 6
//	db b'00001100', b'00001000'; line 7 / 8
//	;; char #2
//	db b'00000000', b'00000000'; line 1 / 2
//	db b'00001110', b'00001110'; line 3 / 4
//	db b'00001110', b'00001110'; line 5 / 6
//	db b'00000000', b'00000000'; line 7 / 8
//	;; char #3
//	db b'00000000', b'00001010'; line 1 / 2
//	db b'00001010', b'00001010'; line 3 / 4
//	db b'00001010', b'00001010'; line 5 / 6
//	db b'00001010', b'00000000'; line 7 / 8

const unsigned char lcd_charset[3*8] = {
  0x08, 0x0c, 0x0e, 0x0f, 0x0f, 0x0e, 0x0c, 0x08, // Play
  0x00, 0x00, 0x0e, 0x0e, 0x0e, 0x0e, 0x00, 0x00, // Stop
  0x00, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x0a, 0x00, // Pause
  };

void DISPLAY_Init(void) __wparam
{
  // init default charset
  MIOS_CLCD_SpecialCharsInit(lcd_charset);

  // clear screen
  MIOS_LCD_Clear();

  // print default screen



  // request display update
  app_flags.DISPLAY_UPDATE_REQ = 1;
}

/////////////////////////////////////////////////////////////////////////////
//  This function is called in the mainloop when no temporary message is shown
//  on screen. Print the realtime messages here
/////////////////////////////////////////////////////////////////////////////
void DISPLAY_Tick(void) __wparam
{


   // Display will look like this:
   //	31.SongName  120   (song number / song name / original BPM of audio loops)
   //   >15.4/16  12 125   (stop/play symbols / current measure & beat / current section / current bpm)

  // do nothing if no update has been requested
  if( !app_flags.DISPLAY_UPDATE_REQ )
    return;

  // clear request
  app_flags.DISPLAY_UPDATE_REQ = 0;

  // Print song number 
  MIOS_LCD_CursorSet(0x00); // first line
  MIOS_LCD_PrintBCD2(MCLOCK_SongNumberGet() +1);


  // Print "." is section or song is about to change on the next measure

  if (mclock_ctr_changesection == 1) {
    MIOS_LCD_PrintChar('.'); 
  } else {
  MIOS_LCD_PrintChar(' '); 

   }
  // Print Song name

  MIOS_LCD_PrintCString( (songname[MCLOCK_SongNumberGet()]) ); 

 
  // Print total number of tracks for this song (only for debugging purposes for now)
//  MIOS_LCD_PrintBCD2(mclock_ctr_loadseq[3]);

  MIOS_LCD_CursorSet(0x00 + 13);

  // Print original BPM of loaded song's loops
  MIOS_LCD_PrintBCD3(mclock_ctr_loadseq[0]);


  // print Play/Stop/Pause char at the lower left corner
  MIOS_LCD_CursorSet(0x40);
  if( mclock_state.PAUSE ) {
    MIOS_LCD_PrintChar(0x02); // Pause
  } else {
    if( mclock_state.RUN ) {
      MIOS_LCD_PrintChar(0x00); // Play
    } else {
      MIOS_LCD_PrintChar(0x01); // Stop
    }
  }


  // print the meter at lower line, left side
  MIOS_LCD_CursorSet(0x40 + 1);
  MIOS_LCD_PrintBCD2(mclock_ctr_measures + 1);
  MIOS_LCD_PrintChar(':');
  MIOS_LCD_PrintBCD1(mclock_ctr_beats + 1);

  MIOS_LCD_CursorSet(0x40 + 5);
  MIOS_LCD_PrintChar('/');

  // print current section length in measures
  MIOS_LCD_PrintBCD2(mclock_ctr_seclength);
  MIOS_LCD_PrintChar(' ');

  // print current section number
  MIOS_LCD_PrintBCD2(mclock_ctr_section +1);
  MIOS_LCD_PrintChar('s ');
  MIOS_LCD_PrintChar(' ');

  // print actual BPM at lower line, right side
  MIOS_LCD_PrintBCD3(MCLOCK_BPMGet());



}

/////////////////////////////////////////////////////////////////////////////
//  This function is called by MIOS when a complete MIDI event has been received
/////////////////////////////////////////////////////////////////////////////
void MPROC_NotifyReceivedEvnt(unsigned char evnt0, unsigned char evnt1, unsigned char evnt2) __wparam
{
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when a MIDI event has been received
// which has been specified in the MIOS_MPROC_EVENT_TABLE
/////////////////////////////////////////////////////////////////////////////
void MPROC_NotifyFoundEvent(unsigned entry, unsigned char evnt0, unsigned char evnt1, unsigned char evnt2) __wparam
{
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when a MIDI event has not been completly
// received within 2 seconds
/////////////////////////////////////////////////////////////////////////////
void MPROC_NotifyTimeout(void) __wparam
{
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when a MIDI byte has been received
/////////////////////////////////////////////////////////////////////////////
void MPROC_NotifyReceivedByte(unsigned char byte) __wparam
{
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS before the shift register are loaded
/////////////////////////////////////////////////////////////////////////////
void SR_Service_Prepare(void) __wparam
{
  // this function is called each millisecond
  // we are using it to output the MIDI clock at Pin D.4 for 1 mS
  if( mclock_pin_state.CLK_REQ ) {
    mclock_pin_state.CLK_REQ = 0;
    PORTDbits.RD4 = 1;
  } else {
    PORTDbits.RD4 = 0;
  }
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS after the shift register have been loaded
/////////////////////////////////////////////////////////////////////////////
void SR_Service_Finish(void) __wparam
{
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when an button has been toggled
// pin_value is 1 when button released, and 0 when button pressed
/////////////////////////////////////////////////////////////////////////////
void DIN_NotifyToggle(unsigned char pin, unsigned char pin_value) __wparam
{
 
  switch( pin ) {
    case 17: // Stop Button
      if( pin_value == 0 ) 
	MCLOCK_DoStop();
      break;

    case 16: // Stutter Button
      if( pin_value == 0 ) 
      MCLOCK_DoStutter();
      break;


    case 0: // Section Mute/Unmute Button
      if( pin_value == 0 ) 

       MCLOCK_DoTrackMute(1);

        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;

    case 1: // Section Mute/Unmute Button
      if( pin_value == 0 ) 

       MCLOCK_DoTrackMute(2);

        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;

    case 2: // Section Mute/Unmute Button
      if( pin_value == 0 ) 

       MCLOCK_DoTrackMute(3);

        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;


    case 9: // Send tempo to Kontakt Script
      if( pin_value == 0 ) 

	 MCLOCK_DoRew();

         MIOS_DOUT_PinSet(21,1); 
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

      if( pin_value == 1 ) 
       MIOS_DOUT_PinSet(21,0); 
       break;


    case 18: // Encoder Button - load selected song
      if( pin_value == 0 ) 
       MCLOCK_DoSongLoad();
       MCLOCK_DoSectionSet(0);
       MIOS_DOUT_PinSet(20,1); 
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

      if( pin_value == 1 ) 
       MIOS_DOUT_PinSet(20,0); 

       break;

    case 7: // Section 1 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(0);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;


    case 6: // Section 2 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(1);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 5: // Section 3 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(2);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;

    case 4: // Section 4 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 

       MCLOCK_DoSectionSet(3);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 15: // Section 5 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 

       MCLOCK_DoSectionSet(4);

        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;


    case 14: // Section 6 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 

       MCLOCK_DoSectionSet(5);

        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;


    case 13: // Section 7 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 

       MCLOCK_DoSectionSet(6);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 12: // Section 8 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(7);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 23: // Section 9 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(8);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;

    case 22: // Section 10 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(9);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;

    case 21: // Section 11 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(10);        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 20: // Section 12 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(11);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 31: // Section 13 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 

       MCLOCK_DoSectionSet(12);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;


    case 30: // Section 14 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(13);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;
       break;



    case 29: // Section 15 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(14);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;



    case 28: // Section 16 Button - change Section value and notify of upcoming section change
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(15);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;


 case 27: // Footswitch = section +1
      if( pin_value == 0 ) 
       MCLOCK_DoSectionSet(mclock_ctr_section+1);
        // request display update
        app_flags.DISPLAY_UPDATE_REQ = 1;

       break;

    case 10: // Down BPM
      if( pin_value == 0 ) {

     unsigned int value;

     value = (unsigned int)MCLOCK_BPMGet() - 48;
       if( MIOS_HLP_16bitAddSaturate(-1, &value, 255 - 48) ) {
       MCLOCK_BPMSet((unsigned char)value + 48);
       app_flags.DISPLAY_UPDATE_REQ = 1;
       }
    }
      break;

    case 11: // Up BPM
      if( pin_value == 0 ) {

     unsigned int value;

     value = (unsigned int)MCLOCK_BPMGet() - 48;
       if( MIOS_HLP_16bitAddSaturate(1, &value, 255 - 48) ) {
       MCLOCK_BPMSet((unsigned char)value + 48);
       app_flags.DISPLAY_UPDATE_REQ = 1;
       }
    }
      break;


  }
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when an encoder has been moved
// incrementer is positive when encoder has been turned clockwise, else
// it is negative
/////////////////////////////////////////////////////////////////////////////

void ENC_NotifyChange(unsigned char encoder, char incrementer) __wparam
{

  unsigned int value;
  incrementer *= -1;

  // Select song number to be loaded 

  if( encoder == 0 ) {
    value = (unsigned int)MCLOCK_SongNumberGet();
    if( MIOS_HLP_16bitAddSaturate(incrementer, &value, 46) ) {
      MCLOCK_SongNumberSet((unsigned char)value);


      app_flags.DISPLAY_UPDATE_REQ = 1;
    }
  }
}






/////////////////////////////////////////////////////////////////////////////
// This function is called by MIOS when a pot has been moved.
// The pots are mapped to midi CC's specified by the pot_event_map array
/////////////////////////////////////////////////////////////////////////////
void AIN_NotifyChange(unsigned char pin, unsigned int pin_value) __wparam
{
  // send mapped CC value
  MIOS_MIDI_TxBufferPut(pot_event_map[pin][0]); // first value from table
  MIOS_MIDI_TxBufferPut(pot_event_map[pin][1]); // second value from table
  MIOS_MIDI_TxBufferPut(MIOS_AIN_Pin7bitGet(pin)); // 7bit pot value
}
